/*
 * yasapi_ring.c
 *
 * Yet Another (WA)SAPI (YASAPI) New Technology (NT) Output (out_yasapi-nt)
 * Copyright (C) 2015-2018  Peter Belkner (pbelkner@snafu.de)
 *
 * out_yasapi-nt is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * out_yasapi-nt is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with out_yasapi-nt.  If not, see <https://www.gnu.org/licenses/>.
 */
#include <yasapi.h>
#include <math.h>

int RingCreate(Ring *pRing, IRingVtbl *lpVtbl, Engine *pEngine, DWORD dwSize)
{
  pRing->lpVtbl=lpVtbl;
  pRing->pEngine=pEngine;

  dwSize=lpVtbl->target.ConvertSize(pRing,dwSize);

  if (!(pRing->_buf=malloc(dwSize))) {
    DMESSAGE("allocation ring");
    goto emalloc;
  }

  pRing->_rp=pRing->_buf;
  pRing->_wp=pRing->_buf;
  pRing->_mp=pRing->_buf+dwSize;
  pRing->_dwWritten=0ul;
  pRing->_dwAvailable=dwSize;
DVWRITELN("dwAvailable=%lu dwWritten=%lu",pRing->_dwAvailable,pRing->_dwWritten);
emalloc:
  return 0;
}

void RingDestroy(Ring *pRing)
{
  if (pRing->_buf) {
    free(pRing->_buf);
    pRing->_buf=NULL;
  }

  pRing->pEngine=NULL;
}

void RingReset(Ring *pRing)
{
  if (!pRing->pEngine) {
    DMESSAGE("ring not initialized");
    goto enoinit;
  }

  EnterCriticalSection(&pRing->pEngine->cs); // {
  pRing->_rp=pRing->_buf;
  pRing->_wp=pRing->_buf;
  pRing->_dwWritten=0ul;
  pRing->_dwAvailable=pRing->_mp-pRing->_buf;
  LeaveCriticalSection(&pRing->pEngine->cs); // }
enoinit:
  return;
}

int RingWrite(Ring *pRing, BYTE *buf, DWORD dwSize)
{
  int code=-1;
  DWORD dwSize1,dwSize2;

  if (!pRing->pEngine) {
    DMESSAGE("ring not initialized");
    goto enoinit;
  }

  EnterCriticalSection(&pRing->pEngine->cs); // {
  dwSize=pRing->lpVtbl->target.ConvertSize(pRing,dwSize);

  if (pRing->_wp<pRing->_rp) {
    // already wrapped
    if (pRing->_rp<pRing->_wp+dwSize) {
      DMESSAGE("overflow");
      goto eoverflow;
    }

    pRing->lpVtbl->CopyToRing(pRing,pRing->_wp,buf,dwSize);
    pRing->_wp+=dwSize;
  }
  else if (pRing->_mp<pRing->_wp+dwSize) {
    // will wrap
    dwSize1=pRing->_mp-pRing->_wp;
    dwSize2=dwSize-dwSize1;

    if (pRing->_rp<pRing->_buf+dwSize2) {
      DMESSAGE("overflow");
      goto eoverflow;
    }

    pRing->lpVtbl->CopyToRing(pRing,pRing->_wp,buf,dwSize1);
    pRing->lpVtbl->CopyToRing(pRing,pRing->_buf,buf+dwSize1,dwSize2);
    pRing->_wp=pRing->_buf+dwSize2;
  }
  else {
    // will not wrap
    if (pRing->_mp<pRing->_wp+dwSize) {
      DMESSAGE("overflow");
      goto eoverflow;
    }

    pRing->lpVtbl->CopyToRing(pRing,pRing->_wp,buf,dwSize);
    pRing->_wp+=dwSize;
  }

  pRing->_dwAvailable-=dwSize;
  pRing->_dwWritten+=dwSize;

  code=0;
eoverflow:
  LeaveCriticalSection(&pRing->pEngine->cs); // }
enoinit:
  return code;
}

int RingRead(Ring *pRing, BYTE *buf, DWORD dwSize)
{
  int code=-1;
  DWORD dwSize1,dwSize2;

  if (!pRing->pEngine) {
    DMESSAGE("ring not initialized");
    goto enoinit;
  }

  EnterCriticalSection(&pRing->pEngine->cs); // {

  if (pRing->_rp<pRing->_wp) {
    // not wrapped
    if (pRing->_wp<pRing->_rp+dwSize) {
      DMESSAGE("underflow");
      goto eunderflow;
    }

    memcpy(buf,pRing->_rp,dwSize);
    pRing->_rp+=dwSize;
  }
  else {
    // wrapped
    dwSize1=pRing->_mp-pRing->_rp;

    if (dwSize<dwSize1) {
      memcpy(buf,pRing->_rp,dwSize);
      pRing->_rp+=dwSize;
    }
    else {
      dwSize2=dwSize-dwSize1;
      memcpy(buf,pRing->_rp,dwSize1);
      memcpy(buf+dwSize1,pRing->_buf,dwSize2);
      pRing->_rp=pRing->_buf+dwSize2;
    }
  }

  pRing->_dwWritten-=dwSize;
  pRing->_dwAvailable+=dwSize;

  code=0;
eunderflow:
  LeaveCriticalSection(&pRing->pEngine->cs); // }
enoinit:
  return code;
}

///////////////////////////////////////////////////////////////////////////////
static DWORD RingPlainConvertSizeSource(Ring *pRing, DWORD dwSize)
{
  return dwSize;
}

static DWORD RingPlainSizeSource(Ring *pRing)
{
  return RingPlainConvertSizeSource(pRing,pRing->_mp-pRing->_buf);
}

static DWORD RingPlainWrittenSource(Ring *pRing)
{
  return RingPlainConvertSizeSource(pRing,pRing->_dwWritten);
}

static DWORD RingPlainAvailableSource(Ring *pRing)
{
  return RingPlainConvertSizeSource(pRing,pRing->_dwAvailable);
}

////////////////
static DWORD RingPlainConvertSizeTarget(Ring *pRing, DWORD dwSize)
{
  return dwSize;
}

static DWORD RingPlainSizeTarget(Ring *pRing)
{
  return pRing->_mp-pRing->_buf;
}

static DWORD RingPlainWrittenTarget(Ring *pRing)
{
  return pRing->_dwWritten;
}

#if defined (YASAPI_TARGET_AVAILABLE) // [
static DWORD RingPlainAvailableTarget(Ring *pRing)
{
  return pRing->_dwAvailable;
}
#endif // ]

////////////////
static BYTE *RingPlainCopyToRing(Ring *pRing, BYTE *dst, const BYTE *src,
    DWORD dwSize)
{
  return memcpy(dst,src,dwSize);
}

static BYTE *RingPlainCopyFromRing(Ring *pRing, BYTE *dst, const BYTE *src,
    DWORD dwSize)
{
  return memcpy(dst,src,dwSize);
}

////////////////
IRingVtbl gcIRingPlainVtbl={
#if defined (YASAPI_SOURCE_CONVERT_SIZE) // [
  .source.ConvertSize=RingPlainConvertSizeSource,
#endif // ]
  .source.Size=RingPlainSizeSource,
  .source.Written=RingPlainWrittenSource,
  .source.Available=RingPlainAvailableSource,
  ////////////
  .target.ConvertSize=RingPlainConvertSizeTarget,
  .target.Size=RingPlainSizeTarget,
  .target.Written=RingPlainWrittenTarget,
#if defined (YASAPI_TARGET_AVAILABLE) // [
  .target.Available=RingPlainAvailableTarget,
#endif // ]
  ////////////
  .CopyToRing=RingPlainCopyToRing,
  .CopyFromRing=RingPlainCopyFromRing,
};

///////////////////////////////////////////////////////////////////////////////
static DWORD RingMono2StereoConvertSizeSource(Ring *pRing, DWORD dwSize)
{
  return dwSize>>1ul;
}

static DWORD RingMono2StereoSizeSource(Ring *pRing)
{
  return RingMono2StereoConvertSizeSource(pRing,pRing->_mp-pRing->_buf);
}

static DWORD RingMono2StereoWrittenSource(Ring *pRing)
{
  return RingMono2StereoConvertSizeSource(pRing,pRing->_dwWritten);
}

static DWORD RingMono2StereoAvailableSource(Ring *pRing)
{
  return RingMono2StereoConvertSizeSource(pRing,pRing->_dwAvailable);
}

////////////////
static DWORD RingMono2StereoConvertSizeTarget(Ring *pRing, DWORD dwSize)
{
  return dwSize<<1ul;
}

static DWORD RingMono2StereoSizeTarget(Ring *pRing)
{
  return pRing->_mp-pRing->_buf;
}

static DWORD RingMono2StereoWrittenTarget(Ring *pRing)
{
  return pRing->_dwWritten;
}

#if defined (YASAPI_TARGET_AVAILABLE) // [
static DWORD RingMono2StereoAvailableTarget(Ring *pRing)
{
  return pRing->_dwAvailable;
}
#endif // ]

////////////////
static double xMono2StereoSqrt2Inverse=0.0;

static BYTE *RingMono2StereoCopyToRing(Ring *pRing, BYTE *pbDst,
    const BYTE *pbSrc, DWORD dwSize)
{
  Engine *pEngine=pRing->pEngine;
#if defined (YASAPI_OPTION_BITDEPTH) // [
  WORD nBlockAlignSource=pEngine->wfx.source.Format.nBlockAlign;
#else // ] [
  WORD nBlockAlignSource=pEngine->params.source.nBlockAlign;
#endif // ]
  BYTE *pbDstMax=pbDst+dwSize;
  INT32 i;
  BYTE *ip=(BYTE *)&i+(sizeof i)-nBlockAlignSource;

  switch (pEngine->options.common.eMono2Stereo) {
  case eMono2StereoHalf:
    while (pbDst<pbDstMax) {
      i=0;
      memcpy(ip,pbSrc,nBlockAlignSource);
      //i=0.5*i+0.5;
      i>>=1;
      memcpy(pbDst,ip,nBlockAlignSource);
      pbDst+=nBlockAlignSource;
      memcpy(pbDst,ip,nBlockAlignSource);
      pbDst+=nBlockAlignSource;
      pbSrc+=nBlockAlignSource;
    }

    break;
  case eMono2StereoSqrt2Inverse:
    if (0.0==xMono2StereoSqrt2Inverse)
      xMono2StereoSqrt2Inverse=1.0/sqrt(2.0);

    while (pbDst<pbDstMax) {
      i=0;
      memcpy(ip,pbSrc,nBlockAlignSource);
      i=xMono2StereoSqrt2Inverse*i+0.5;
      memcpy(pbDst,ip,nBlockAlignSource);
      pbDst+=nBlockAlignSource;
      memcpy(pbDst,ip,nBlockAlignSource);
      pbDst+=nBlockAlignSource;
      pbSrc+=nBlockAlignSource;
    }

    break;
  default:
    while (pbDst<pbDstMax) {
      memcpy(pbDst,pbSrc,nBlockAlignSource);
      pbDst+=nBlockAlignSource;
      memcpy(pbDst,pbSrc,nBlockAlignSource);
      pbDst+=nBlockAlignSource;
      pbSrc+=nBlockAlignSource;
    }

    break;
  }

  return pbDst;
}

static BYTE *RingMono2StereoCopyFromRing(Ring *pRing,
    BYTE *pbDst, const BYTE *pbSrc, DWORD dwSize)
{
  return memcpy(pbDst,pbSrc,dwSize);
}

////////////////
IRingVtbl gcIRingMono2StereoVtbl={
#if defined (YASAPI_SOURCE_CONVERT_SIZE) // [
  .source.ConvertSize=RingMono2StereoConvertSizeSource,
#endif // ]
  .source.Size=RingMono2StereoSizeSource,
  .source.Written=RingMono2StereoWrittenSource,
  .source.Available=RingMono2StereoAvailableSource,
  ////////////
  .target.ConvertSize=RingMono2StereoConvertSizeTarget,
  .target.Size=RingMono2StereoSizeTarget,
  .target.Written=RingMono2StereoWrittenTarget,
#if defined (YASAPI_TARGET_AVAILABLE) // [
  .target.Available=RingMono2StereoAvailableTarget,
#endif // ]
  ////////////
  .CopyToRing=RingMono2StereoCopyToRing,
  .CopyFromRing=RingMono2StereoCopyFromRing,
};
